﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ArcGIS.Core.CIM;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Mapping;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class MapObjectLayer manages the display of a list of MapObjects as a layer in ArcGIS
    /// </summary>
    class MapObjectLayer
    {
        /** The name of the layer */
        private string name;
        /** The host of the 3D Mapping viewer */
        private PluginArcGIS viewerHost;
        /** The map objects */
        private AList<MapObject> mapObjects;
        /** Do we need a refresh? */
        private bool mapRefresh;
        /** The ArcGIS map object overlays */
        private List<MapObjectOverlay> mapOverlays;
        /* Visible? */
        private bool visible;
        /** Closed? */
        private bool closed;

        /// <summary>
        /// Create a new layer
        /// </summary>
        /// <param name="viewerHost">the host of the 3D Mapping viewer</param>
        internal MapObjectLayer(string name,PluginArcGIS viewerHost)
        {
            this.name = name;
            this.viewerHost = viewerHost;
            this.mapObjects = new AList<MapObject>(MapObject.TYPE);
            this.mapRefresh = false;
            this.mapOverlays = new List<MapObjectOverlay>();
            this.visible = true;
            this.closed = false;
            /* Request a map update */
            this.viewerHost.RequestMapUpdate("added mapobject layer '"+this.name+"'");
        }

        /// <summary>
        /// Change the visibility of the layer
        /// </summary>
        /// <param name="visible"></param>
        internal void setVisible(bool visible)
        {
            this.visible = visible;
            this.mapRefresh = true;
        }

        /// <summary>
        /// Update the layer to display a new list of map objects
        /// Calling this method does not trigger a map-view update, the caller should trigger one
        /// </summary>
        internal void update(AList<MapObject> mapObjects)
        {
            Message.WriteLine("updating mapobject layer '" + this.name + "'");
            if (mapObjects == null) return;
            this.mapObjects = mapObjects;
            this.mapRefresh = true;
        }

        /// <summary>
        /// Unload the layer
        /// </summary>
        internal void unload()
        {
            /* Just clear the list of map objects */
            this.update(new AList<MapObject>(MapObject.TYPE));
            /* Request a map update */
            this.viewerHost.RequestMapUpdate("unloaded mapobject layer '" + this.name + "'");
        }

        /// <summary>
        /// Update the map-view for this layer. This method is called by the plugin host from the UI thread (MCT).
        /// </summary>
        /// <param name="mapView">the map-view to display on</param>
        internal void updateMapView(MapView mapView)
        {
            /* No need to update? */
            if (this.mapRefresh == false) return;
            this.mapRefresh = false;
            Message.WriteLine("updating mapobject layer '" + this.name + "'");
            /* Closed? */
            if (this.closed)
            {
                /* Clear the overlays */
                foreach (MapObjectOverlay mapOverlay in this.mapOverlays) mapOverlay.Close(false/*recycle*/);
                this.mapOverlays.Clear();
                /* Reset */
                this.mapObjects = new AList<MapObject>(MapObject.TYPE);
                this.mapRefresh = false;
                return;
            }
            /* Invisible? */
            if (this.visible==false)
            {
                /* Clear the overlays */
                foreach (MapObjectOverlay mapOverlay in this.mapOverlays) mapOverlay.Close(false/*recycle*/);
                this.mapOverlays.Clear();
                return;
            }
            /* Clear the overlays */
            List<System.IDisposable> recycledOverlays = new List<System.IDisposable>();
            foreach (MapObjectOverlay mapOverlay in this.mapOverlays)
            {
                /* Clear the next overlay and recycle to avoid flickering */
                System.IDisposable recycledOverlay = mapOverlay.Close(true/*recycle*/);
                if (recycledOverlay != null) recycledOverlays.Add(recycledOverlay);
            }
            this.mapOverlays.Clear();
            /* Create new overlays */
            for (int i = 0; i < this.mapObjects.size(); i++)
            {
                /* Make a new overlay */
                MapObject mapObject = this.mapObjects.get(i);
                System.IDisposable recycledOverlay = (i < recycledOverlays.Count) ? recycledOverlays[i] : null;
                MapObjectOverlay mapOverlay = new MapObjectOverlay(mapObject, mapView, true/*addToMapView*/, recycledOverlay);
                this.mapOverlays.Add(mapOverlay);
            }
            /* Remove unused overlays */
            for (int i = mapObjects.size(); i < recycledOverlays.Count; i++) recycledOverlays[i].Dispose();
        }
    }
}
