﻿using System;
using System.Diagnostics;
using System.Threading.Tasks;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class UpdateFieldOfView manages the updates to the field-of-view layer
    /// </summary>
    class UpdateFieldOfView
    {
        private MapObjectLayer layer;
        private PluginArcGIS viewerHost;
        private long viewsUpdateTime;
        private long lastRequestTime;
        private bool isUpdating;
        private AList<MapObject> mapObjects;

        /// <summary>
        /// Create a new updater
        /// </summary>
        internal UpdateFieldOfView(MapObjectLayer layer, PluginArcGIS viewerHost)
        {
            this.layer = layer;
            this.viewerHost = viewerHost;
            this.viewsUpdateTime = 0;
            this.lastRequestTime = 0;
            this.isUpdating = false;
            this.mapObjects = null;
        }

        /// <summary>
        /// Called when the views have changed and the field-of-views need to be reloaded
        /// </summary>
        internal void updatedViews()
        {
            this.viewsUpdateTime = DateTime.Now.Ticks;
        }

        /// <summary>
        /// Check for updates
        /// </summary>
        internal void checkForUpdates()
        {
            // already updating?
            if (this.isUpdating) return;
            // already up-to-date?
            if (this.lastRequestTime > this.viewsUpdateTime) return;
            long currentTime = DateTime.Now.Ticks;
            // throttle the rate of update requests
            if ((this.lastRequestTime > 0) && (currentTime < this.lastRequestTime + 200 * TimeSpan.TicksPerMillisecond)) return;
            // throttle the rate of view changes (fast panning)
            if ((this.viewsUpdateTime > 0) && (currentTime < this.viewsUpdateTime + 200 * TimeSpan.TicksPerMillisecond)) return;
            // start the update task (to free the calling timer thread)
            this.isUpdating = true;
            this.lastRequestTime = currentTime;
            Task.Run(() => this.update());
        }

        /// <summary>
        /// Update
        /// </summary>
        private void update()
        {
            try
            {
                Message.WriteLine("Requesting field-of-views");
                this.mapObjects = this.viewerHost.getViewer().getFieldOfViews2D();
                Message.WriteLine("Got " + this.mapObjects.size() + " mapobjects");
                this.isUpdating = false;
                /* Update the layer */
                this.layer.update(this.mapObjects);
                this.viewerHost.RequestMapUpdate("updated field-of-views");
                return;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while updating field-of-views: " + e);
                this.isUpdating = false;
                return;
            }
        }
    }
}
